/**
 * Back 2 Browser Bytecode Translator
 * Copyright (C) 2012 Jaroslav Tulach <jaroslav.tulach@apidesign.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. Look for COPYING file in the top folder.
 * If not, see http://opensource.org/licenses/GPL-2.0.
 */
package org.apidesign.bck2brwsr.htmlpage;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.reflect.Method;
import java.util.logging.Level;
import java.util.logging.Logger;
import javafx.scene.web.WebEngine;
import netscape.javascript.JSObject;
import org.apidesign.bck2brwsr.core.ExtraJavaScript;
import org.apidesign.bck2brwsr.core.JavaScriptBody;

/** Provides binding between models and 
 *
 * @author Jaroslav Tulach <jtulach@netbeans.org>
 */
@ExtraJavaScript(resource = "/org/apidesign/bck2brwsr/htmlpage/knockout-2.2.1.js")
public class Knockout {
    private static final Logger LOG = Logger.getLogger(Knockout.class.getName());
    /** used by tests */
    static Knockout next;
    
    static {
        BufferedReader r = new BufferedReader(new InputStreamReader(Knockout.class.getResourceAsStream("knockout-2.2.1.js")));
        StringBuilder sb = new StringBuilder();
        for (;;) {
            try {
                String l = r.readLine();
                if (l == null) {
                    break;
                }
                sb.append(l).append('\n');
            } catch (IOException ex) {
                throw new IllegalStateException(ex);
            }
        }
        web().executeScript(sb.toString());
        Object ko = web().executeScript("ko");
        assert ko != null : "Knockout library successfully defined 'ko'";
    }
    
    
    private final Object bindings;
    
    Knockout(Object bindings) {
        this.bindings = bindings;
    }
    
    public static <M> Knockout applyBindings(
        Class<M> modelClass, M model, String[] propsGettersAndSetters
    ) {
        Object bindings = next;
        next = null;
        if (bindings == null) {
            bindings = web().executeScript("new Object()");
        }
        for (int i = 0; i < propsGettersAndSetters.length; i += 4) {
            try {
                Method getter = modelClass.getMethod(propsGettersAndSetters[i + 3]);
                bind(bindings, model, propsGettersAndSetters[i],
                    propsGettersAndSetters[i + 1],
                    propsGettersAndSetters[i + 2],
                    getter.getReturnType().isPrimitive()
                );
            } catch (NoSuchMethodException ex) {
                throw new IllegalStateException(ex.getMessage());
            }
        }
        applyBindings(bindings);
        return new Knockout(bindings);
    }

    @JavaScriptBody(args = { "prop" }, body =
        "this[prop].valueHasMutated();"
    )
    public void valueHasMutated(String prop) {
        LOG.log(Level.FINE, "property mutated: {0}", prop);
        JSObject koProp = (JSObject) ((JSObject)bindings).getMember(prop);
        koProp.call("valueHasMutated");
    }
    

    @JavaScriptBody(args = { "id", "ev" }, body = "ko.utils.triggerEvent(window.document.getElementById(id), ev.substring(2));")
    public static void triggerEvent(String id, String ev) {
    }
    
    @JavaScriptBody(args = { "bindings", "model", "prop", "getter", "setter", "primitive" }, body =
          "var bnd = {\n"
        + "  read: function() {\n"
        + "    var v = model[getter]();\n"
        + "    return v;\n"
        + "  },\n"
        + "  owner: bindings\n"
        + "};\n"
        + "if (setter != null) {\n"
        + "  bnd.write = function(val) {\n"
        + "    model[setter](primitive ? new Number(val) : val);\n"
        + "  };\n"
        + "}\n"
        + "bindings[prop] = ko.computed(bnd);"
    )
    private static void bind(
        Object bindings, Object model, String prop, String getter, String setter, boolean primitive
    ) {
        WebEngine e = web();
        JSObject bnd = (JSObject) e.executeScript("var x = {}; x.bnd = "
        + "new Function('ko', 'bindings', 'model', 'prop', 'getter', 'setter', 'primitive', '"
        + "var bnd = {"
        + "  read: function() {"
        + "    try {"
        + "      return model[getter]();"
        + "    } catch (e) {"
        + "      alert(\"Cannot call \" + getter + \" on \" + model + \" error: \" + e);"
        + "    }"
        + "  },"
        + "  owner: bindings"
        + "};"
        + "if (setter != null) {"
        + "  bnd.write = function(val) {"
        + "    model[setter](primitive ? new Number(val) : val);"
        + "  };"
        + "};"
        + "bindings[prop] = ko.computed(bnd);'"
        + "); x;");
        
        Object ko = e.executeScript("ko");
        bnd.call("bnd", ko, bindings, model, prop, strip(getter), strip(setter), primitive);
        LOG.log(Level.FINE, "binding defined for {0}: {1}", new Object[]{prop, ((JSObject)bindings).getMember(prop)});
    }
    
    private static String strip(String mangled) {
        if (mangled == null) {
            return null;
        }
        int under = mangled.indexOf("__");
        return mangled.substring(0, under);
    }
    
    @JavaScriptBody(args = { "bindings" }, body = "ko.applyBindings(bindings);")
    private static void applyBindings(Object bindings) {
        JSObject ko = (JSObject) web().executeScript("ko");
        ko.call("applyBindings", bindings);
    }
    
    private static WebEngine web() {
        return (WebEngine) System.getProperties().get("webEngine");
    }
    
}
